// -------------------------------------------------
// Copyright (c) 2022 HiBit <https://www.hibit.dev>
// -------------------------------------------------

#include "Wire.h"
#include "I2C.h"

// ---- Code for motors ----
#include <Servo.h>

// ---- Code for motors done ----

// ---- Code for PID ----
#include "PID_v1.h"
// ---- Code for PID done ----

#define LEFT 11
#define RIGHT 12
#define MPU9250_IMU_ADDRESS 0x68
#define MPU9250_MAG_ADDRESS 0x0C

#define GYRO_FULL_SCALE_250_DPS  0x00
#define GYRO_FULL_SCALE_500_DPS  0x08
#define GYRO_FULL_SCALE_1000_DPS 0x10
#define GYRO_FULL_SCALE_2000_DPS 0x18

#define ACC_FULL_SCALE_2G  0x00
#define ACC_FULL_SCALE_4G  0x08
#define ACC_FULL_SCALE_8G  0x10
#define ACC_FULL_SCALE_16G 0x18

#define TEMPERATURE_OFFSET 21 // As defined in documentation

#define INTERVAL_MS_PRINT 1000

#define G 9.80665

struct gyroscope_raw {
  int16_t x, y, z;
} gyroscope;

struct {
  struct {
    float x, y, z;
  } accelerometer, gyroscope, magnetometer;

  float temperature;
} normalized;

unsigned long lastPrintMillis = 0;

// -------------------------------------------------
// Copyright (c) 2022 HiBit <https://www.hibit.dev>
// -------------------------------------------------


bool isImuReady()
{
  uint8_t isReady; // Interruption flag

  I2Cread(MPU9250_IMU_ADDRESS, 58, 1, &isReady);

  return isReady & 0x01; // Read register and wait for the RAW_DATA_RDY_INT
}

void readRawImu()
{
  uint8_t buff[14];

  // Read output registers:
  // [59-64] Accelerometer
  // [65-66] Temperature
  // [67-72] Gyroscope
  I2Cread(MPU9250_IMU_ADDRESS, 59, 14, buff);

  // Gyroscope, create 16 bits values from 8 bits data
  gyroscope.x = (buff[8] << 8 | buff[9]);
  gyroscope.y = (buff[10] << 8 | buff[11]);
  gyroscope.z = (buff[12] << 8 | buff[13]);
}
void normalize(gyroscope_raw gyroscope)
{
  // Sensitivity Scale Factor (MPU datasheet page 8)
  normalized.gyroscope.x = gyroscope.x / 32.8;
  normalized.gyroscope.y = gyroscope.y / 32.8;
  normalized.gyroscope.z = gyroscope.z / 32.8;
}

// ---- Code for motors ----
Servo servo1;
Servo servo2;
int pos = 90;

void moveMotors(int speedPercentage) {
  float idealPos = 0.9f * speedPercentage + 90;
  pos = (int)idealPos;
  servo1.write(pos);
  servo2.write(180 - pos);
}

void moveMotorsDegree(int degree) {

  pos = 90 + degree;
  if (pos < 45)
    pos = 45;
  if (pos > 135)
    pos = 135;
  Serial.println(pos);
  servo1.write(pos);
  servo2.write(182 - pos);
  //  delay(10);
}

int lastOutput1 = 92;
int lastOutput2 = 91;

void moveMotorsPID(int output) {
  int output1 = 92 - output;
  int output2 = 91 + output; // if robot tends to go to the right, decrease this

  // Limit the change in output
  if (output1 > lastOutput1 + 1) output1 = lastOutput1 + 1;
  if (output1 < lastOutput1 - 1) output1 = lastOutput1 - 1;
  if (output2 > lastOutput2 + 1) output2 = lastOutput2 + 1;
  if (output2 < lastOutput2 - 1) output2 = lastOutput2 - 1;

  servo1.write(output1); // offset from the left servo
  servo2.write(output2); // offset from the right servo

  lastOutput1 = output1;
  lastOutput2 = output2;
}

// ---- Code for motors done ----

// ---- Code for PID ----
// PID parameters
// (these will need to be adjusted to suit your specific robot)
double Kp = 0.29; // 0.29 - ideal
double Ki = 0.11; // 0.11 - ideal
double Kd = 0.011; // 0.01 - ideal

// PID setpoint (the desired angle)
double setpoint = -1.7; // -1.7 ideal

// PID input (the current angle) and output (the motor speed adjustment)
double input, output;

// Initialize the PID controller
PID myPID(&input, &output, &setpoint, Kp, Ki, Kd, DIRECT);
// ---- Code for PID done ----

void setup()
{
  // ---- Code for motors ----
  servo1.attach(LEFT);
  servo2.attach(RIGHT);
  moveMotors(0);
  // ---- Code for motors done ----

  Wire.begin();
  Serial.begin(115200);

  I2CwriteByte(MPU9250_IMU_ADDRESS, 27, GYRO_FULL_SCALE_1000_DPS); // Configure gyroscope range
  I2CwriteByte(MPU9250_IMU_ADDRESS, 28, ACC_FULL_SCALE_2G);        // Configure accelerometer range

  I2CwriteByte(MPU9250_IMU_ADDRESS, 55, 0x02); // Set by pass mode for the magnetometers
  I2CwriteByte(MPU9250_IMU_ADDRESS, 56, 0x01); // Enable interrupt pin for raw data

  //Start magnetometer
  I2CwriteByte(MPU9250_MAG_ADDRESS, 0x0A, 0x12); // Request continuous magnetometer measurements in 16 bits (mode 1)
  Serial.println("starting");

  // Initialize the PID controller
  myPID.SetMode(AUTOMATIC);
  myPID.SetOutputLimits(-45, 45 );  // This should keep the motor input within the valid range
}

void loop()
{

  if (isImuReady()) {
    readRawImu();

    normalize(gyroscope);

    // Get the current angle from the gyroscope
    input = normalized.gyroscope.x;
    Serial.println(input);

    // Calculate the PID adjustment
    myPID.Compute();

    // Apply the PID adjustment to the servo motors
    moveMotorsPID(output);
  }

}
